# Auto generated Python source file from Ada sources
# Call 'make' in 'src/vhdl' to regenerate:
#
from pydecor import export


@export
class Tok:
    Invalid = 0
    Eof = 1
    Newline = 2
    Block_Comment_Start = 3
    Block_Comment_End = 4
    Block_Comment_Text = 5
    Line_Comment = 6
    Character = 7
    Identifier = 8
    Integer = 9
    Real = 10
    String = 11
    Bit_String = 12
    Integer_Letter = 13
    Left_Paren = 14
    Right_Paren = 15
    Left_Bracket = 16
    Right_Bracket = 17
    Colon = 18
    Semi_Colon = 19
    Comma = 20
    Double_Arrow = 21
    Tick = 22
    Double_Star = 23
    Assign = 24
    Bar = 25
    Box = 26
    Dot = 27
    Equal_Equal = 28
    Equal = 29
    Not_Equal = 30
    Less = 31
    Less_Equal = 32
    Greater = 33
    Greater_Equal = 34
    Match_Equal = 35
    Match_Not_Equal = 36
    Match_Less = 37
    Match_Less_Equal = 38
    Match_Greater = 39
    Match_Greater_Equal = 40
    Plus = 41
    Minus = 42
    Ampersand = 43
    Condition = 44
    Double_Less = 45
    Double_Greater = 46
    Caret = 47
    And_And = 48
    Bar_Bar = 49
    Left_Curly = 50
    Right_Curly = 51
    Exclam_Mark = 52
    Brack_Star = 53
    Brack_Plus_Brack = 54
    Brack_Arrow = 55
    Brack_Equal = 56
    Bar_Arrow = 57
    Bar_Double_Arrow = 58
    Minus_Greater = 59
    Equiv_Arrow = 60
    Arobase = 61
    Star = 62
    Slash = 63
    Mod = 64
    Rem = 65
    Abs = 66
    Not = 67
    Access = 68
    After = 69
    Alias = 70
    All = 71
    Architecture = 72
    Array = 73
    Assert = 74
    Attribute = 75
    Begin = 76
    Block = 77
    Body = 78
    Buffer = 79
    Bus = 80
    Case = 81
    Component = 82
    Configuration = 83
    Constant = 84
    Disconnect = 85
    Downto = 86
    Else = 87
    Elsif = 88
    End = 89
    Entity = 90
    Exit = 91
    File = 92
    For = 93
    Function = 94
    Generate = 95
    Generic = 96
    Guarded = 97
    If = 98
    In = 99
    Inout = 100
    Is = 101
    Label = 102
    Library = 103
    Linkage = 104
    Loop = 105
    Map = 106
    New = 107
    Next = 108
    Null = 109
    Of = 110
    On = 111
    Open = 112
    Others = 113
    Out = 114
    Package = 115
    Port = 116
    Procedure = 117
    Process = 118
    Range = 119
    Record = 120
    Register = 121
    Report = 122
    Return = 123
    Select = 124
    Severity = 125
    Signal = 126
    Subtype = 127
    Then = 128
    To = 129
    Transport = 130
    Type = 131
    Units = 132
    Until = 133
    Use = 134
    Variable = 135
    Wait = 136
    When = 137
    While = 138
    With = 139
    And = 140
    Or = 141
    Xor = 142
    Nand = 143
    Nor = 144
    Xnor = 145
    Group = 146
    Impure = 147
    Inertial = 148
    Literal = 149
    Postponed = 150
    Pure = 151
    Reject = 152
    Shared = 153
    Unaffected = 154
    Sll = 155
    Sla = 156
    Sra = 157
    Srl = 158
    Rol = 159
    Ror = 160
    Protected = 161
    Assume = 162
    Context = 163
    Cover = 164
    Default = 165
    Force = 166
    Parameter = 167
    Property = 168
    Release = 169
    Restrict = 170
    Restrict_Guarantee = 171
    Sequence = 172
    Vmode = 173
    Vprop = 174
    Vunit = 175
    Across = 176
    Break = 177
    Limit = 178
    Nature = 179
    Noise = 180
    Procedural = 181
    Quantity = 182
    Reference = 183
    Spectrum = 184
    Subnature = 185
    Terminal = 186
    Through = 187
    Tolerance = 188
    Psl_Clock = 189
    Psl_Endpoint = 190
    Psl_Const = 191
    Psl_Boolean = 192
    Inf = 193
    Within = 194
    Abort = 195
    Before = 196
    Before_Em = 197
    Before_Un = 198
    Before_Em_Un = 199
    Always = 200
    Never = 201
    Eventually_Em = 202
    Next_Em = 203
    Next_A = 204
    Next_A_Em = 205
    Next_E = 206
    Next_E_Em = 207
    Next_Event = 208
    Next_Event_Em = 209
    Next_Event_A = 210
    Next_Event_A_Em = 211
    Next_Event_E = 212
    Next_Event_E_Em = 213
    Until_Em = 214
    Until_Un = 215
    Until_Em_Un = 216
    Prev = 217
    Stable = 218
    Fell = 219
    Rose = 220



import pyGHDL.libghdl.name_table as name_table
import pyGHDL.libghdl.files_map as files_map
import pyGHDL.libghdl.vhdl.nodes as nodes
import pyGHDL.libghdl.vhdl.nodes_meta as nodes_meta
import pyGHDL.libghdl.vhdl.elocations as elocations
import pyGHDL.libghdl.utils as pyutils

from . import lsp

SYMBOLS_MAP = {
    nodes.Iir_Kind.Package_Declaration: {
        "kind": lsp.SymbolKind.Package,
        "detail": "(declaration)",
    },
    nodes.Iir_Kind.Package_Body: {"kind": lsp.SymbolKind.Package, "detail": "(body)"},
    nodes.Iir_Kind.Entity_Declaration: {"kind": lsp.SymbolKind.Module},
    nodes.Iir_Kind.Architecture_Body: {"kind": lsp.SymbolKind.Module},
    nodes.Iir_Kind.Configuration_Declaration: {"kind": lsp.SymbolKind.Module},
    nodes.Iir_Kind.Package_Instantiation_Declaration: {"kind": lsp.SymbolKind.Module},
    nodes.Iir_Kind.Component_Declaration: {"kind": lsp.SymbolKind.Module},
    nodes.Iir_Kind.Context_Declaration: {"kind": lsp.SymbolKind.Module},
    nodes.Iir_Kind.Use_Clause: {"kind": None},
    nodes.Iir_Kind.Library_Clause: {"kind": None},
    nodes.Iir_Kind.Procedure_Declaration: {"kind": lsp.SymbolKind.Function},
    nodes.Iir_Kind.Function_Declaration: {"kind": lsp.SymbolKind.Function},
    nodes.Iir_Kind.Interface_Procedure_Declaration: {"kind": lsp.SymbolKind.Function},
    nodes.Iir_Kind.Interface_Function_Declaration: {"kind": lsp.SymbolKind.Function},
    nodes.Iir_Kind.Procedure_Body: {
        "kind": lsp.SymbolKind.Function,
        "detail": "(body)",
    },
    nodes.Iir_Kind.Function_Body: {"kind": lsp.SymbolKind.Function, "detail": "(body)"},
    nodes.Iir_Kind.Type_Declaration: {"kind": lsp.SymbolKind.Constructor},
    nodes.Iir_Kind.Subtype_Declaration: {"kind": lsp.SymbolKind.Constructor},
    nodes.Iir_Kind.Attribute_Declaration: {"kind": lsp.SymbolKind.Property},
    nodes.Iir_Kind.Attribute_Specification: {"kind": None},
    nodes.Iir_Kind.Disconnection_Specification: {"kind": None},
    nodes.Iir_Kind.Anonymous_Type_Declaration: {"kind": None},
    nodes.Iir_Kind.Variable_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Constant_Declaration: {"kind": lsp.SymbolKind.Constant},
    nodes.Iir_Kind.Signal_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Signal_Attribute_Declaration: {"kind": None},
    nodes.Iir_Kind.File_Declaration: {"kind": lsp.SymbolKind.File},
    nodes.Iir_Kind.Interface_Variable_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Interface_Constant_Declaration: {"kind": lsp.SymbolKind.Constant},
    nodes.Iir_Kind.Interface_Signal_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Interface_File_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.File_Declaration: {"kind": lsp.SymbolKind.File},
    nodes.Iir_Kind.Object_Alias_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Non_Object_Alias_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Protected_Type_Body: {"kind": lsp.SymbolKind.Class},
    nodes.Iir_Kind.Group_Template_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Group_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Concurrent_Simple_Signal_Assignment: {"kind": None},
    nodes.Iir_Kind.Concurrent_Conditional_Signal_Assignment: {"kind": None},
    nodes.Iir_Kind.Concurrent_Selected_Signal_Assignment: {"kind": None},
    nodes.Iir_Kind.Concurrent_Procedure_Call_Statement: {"kind": None},
    nodes.Iir_Kind.Concurrent_Assertion_Statement: {"kind": None},
    nodes.Iir_Kind.Component_Instantiation_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Block_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.If_Generate_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.For_Generate_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Case_Generate_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Sensitized_Process_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Process_Statement: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Psl_Assert_Directive: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Psl_Assume_Directive: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Psl_Cover_Directive: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Psl_Restrict_Directive: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Psl_Endpoint_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Psl_Declaration: {"kind": lsp.SymbolKind.Variable},
    nodes.Iir_Kind.Psl_Assert_Directive: {"kind": lsp.SymbolKind.Method},
    nodes.Iir_Kind.Configuration_Specification: {"kind": None},
}


def location_to_position(fe, loc):
    assert loc != files_map.No_Location
    line = files_map.Location_File_To_Line(loc, fe)
    off = files_map.Location_File_Line_To_Offset(loc, fe, line)
    return {"line": line - 1, "character": off}


def get_symbols_chain(fe, n):
    res = [get_symbols(fe, el) for el in pyutils.chain_iter(n)]
    return [e for e in res if e is not None]


def get_symbols(fe, n):
    if n == nodes.Null_Iir:
        return None
    k = nodes.Get_Kind(n)
    if k == nodes.Iir_Kind.Design_Unit:
        return get_symbols(fe, nodes.Get_Library_Unit(n))
    m = SYMBOLS_MAP.get(k, None)
    if m is None:
        raise AssertionError("get_symbol: unhandled {}".format(pyutils.kind_image(k)))
    kind = m["kind"]
    if kind is None:
        return None
    if k in [nodes.Iir_Kind.Procedure_Declaration, nodes.Iir_Kind.Function_Declaration]:
        # Discard implicit declarations.
        if nodes.Get_Implicit_Definition(n) < nodes.Iir_Predefined.PNone:
            return None
        if nodes.Get_Has_Body(n):
            # Use the body instead.
            # FIXME: but get interface from the spec!
            return None
    res = {"kind": kind}
    detail = m.get("detail")
    if detail is not None:
        res["detail"] = detail
    # Get the name
    if k in [nodes.Iir_Kind.Function_Body, nodes.Iir_Kind.Procedure_Body]:
        nid = nodes.Get_Identifier(nodes.Get_Subprogram_Specification(n))
    else:
        nid = nodes.Get_Identifier(n)
    if nid == name_table.Null_Identifier:
        name = None
    else:
        name = pyutils.name_image(nid)
    # Get the range.  Use elocations when possible.
    if k in (
        nodes.Iir_Kind.Architecture_Body,
        nodes.Iir_Kind.Entity_Declaration,
        nodes.Iir_Kind.Package_Declaration,
        nodes.Iir_Kind.Package_Body,
        nodes.Iir_Kind.Component_Declaration,
        nodes.Iir_Kind.Process_Statement,
        nodes.Iir_Kind.Sensitized_Process_Statement,
        nodes.Iir_Kind.If_Generate_Statement,
        nodes.Iir_Kind.For_Generate_Statement,
    ):
        start_loc = elocations.Get_Start_Location(n)
        end_loc = elocations.Get_End_Location(n)
        if end_loc == files_map.No_Location:
            # Can happen in case of parse error
            end_loc = start_loc
    else:
        start_loc = nodes.Get_Location(n)
        end_loc = start_loc + name_table.Get_Name_Length(nid)
    res["range"] = {
        "start": location_to_position(fe, start_loc),
        "end": location_to_position(fe, end_loc),
    }

    # Gather children.
    # FIXME: should we use a list of fields to inspect ?
    children = []
    # if nodes_meta.Has_Generic_Chain(k):
    #    children.extend(get_symbols_chain(fe, nodes.Get_Generic_Chain(n)))
    # if nodes_meta.Has_Port_Chain(k):
    #    children.extend(get_symbols_chain(fe, nodes.Get_Port_Chain(n)))
    # if nodes_meta.Has_Interface_Declaration_Chain(k):
    #    children.extend(get_symbols_chain(fe, nodes.Get_Interface_Declaration_Chain(n)))
    if k in (nodes.Iir_Kind.Package_Declaration, nodes.Iir_Kind.Package_Body):
        children.extend(get_symbols_chain(fe, nodes.Get_Declaration_Chain(n)))
    if nodes_meta.Has_Concurrent_Statement_Chain(k):
        children.extend(get_symbols_chain(fe, nodes.Get_Concurrent_Statement_Chain(n)))
    if nodes_meta.Has_Generate_Statement_Body(k):
        children.extend(
            get_symbols_chain(
                fe,
                nodes.Get_Concurrent_Statement_Chain(
                    nodes.Get_Generate_Statement_Body(n)
                ),
            )
        )

    if children:
        res["children"] = children
    else:
        # Discard anonymous symbols without children.
        if name is None:
            return None
    res["name"] = name if name is not None else "<anon>"
    return res



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
#  Authors:
#    Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from typing import Callable, List

from pydecor import export


@export
def EnumLookupTable(cls) -> Callable:
	"""
	Decorator to precalculate a enum lookup table (LUT) for enum position to
	enum literal name.

	.. todo:: Make compatible to chained decorators

	:param cls: Enumerator class for which a LUT shall be pre-calculated.
	"""
	def decorator(func) -> Callable:
		def gen() -> List[str]:
			d = [e for e in dir(cls) if e[0] != "_"]
			res = [None] * len(d)
			for e in d:
				res[getattr(cls, e)] = e
			return res

		__lut = gen()

		def wrapper(id: int) -> str:
			# function that replaces the placeholder function
			return __lut[id]

		return wrapper

	return decorator



# Auto generated Python source file from Ada sources
# Call 'make' in 'src/vhdl' to regenerate:
#
from pydecor import export


@export
class Name:
    First_Character = 1
    Last_Character = 256
    First_Keyword = 257
    Mod = 257
    Rem = 258
    Abs = 259
    Not = 260
    Access = 261
    After = 262
    Alias = 263
    All = 264
    Architecture = 265
    Array = 266
    Assert = 267
    Attribute = 268
    Begin = 269
    Block = 270
    Body = 271
    Buffer = 272
    Bus = 273
    Case = 274
    Component = 275
    Configuration = 276
    Constant = 277
    Disconnect = 278
    Downto = 279
    Else = 280
    Elsif = 281
    End = 282
    Entity = 283
    Exit = 284
    File = 285
    For = 286
    Function = 287
    Generate = 288
    Generic = 289
    Guarded = 290
    If = 291
    In = 292
    Inout = 293
    Is = 294
    Label = 295
    Library = 296
    Linkage = 297
    Loop = 298
    Map = 299
    New = 300
    Next = 301
    Null = 302
    Of = 303
    On = 304
    Open = 305
    Others = 306
    Out = 307
    Package = 308
    Port = 309
    Procedure = 310
    Process = 311
    Range = 312
    Record = 313
    Register = 314
    Report = 315
    Return = 316
    Select = 317
    Severity = 318
    Signal = 319
    Subtype = 320
    Then = 321
    To = 322
    Transport = 323
    Type = 324
    Units = 325
    Until = 326
    Use = 327
    Variable = 328
    Wait = 329
    When = 330
    While = 331
    With = 332
    And = 333
    Or = 334
    Xor = 335
    Nand = 336
    Nor = 337
    Last_Vhdl87 = 337
    Xnor = 338
    Group = 339
    Impure = 340
    Inertial = 341
    Literal = 342
    Postponed = 343
    Pure = 344
    Reject = 345
    Shared = 346
    Unaffected = 347
    Sll = 348
    Sla = 349
    Sra = 350
    Srl = 351
    Rol = 352
    Ror = 353
    Last_Vhdl93 = 353
    Protected = 354
    Last_Vhdl00 = 354
    Assume = 355
    Context = 356
    Cover = 357
    Default = 358
    Force = 359
    Parameter = 360
    Property = 361
    Release = 362
    Restrict = 363
    Restrict_Guarantee = 364
    Sequence = 365
    Vmode = 366
    Vprop = 367
    Vunit = 368
    Last_Vhdl08 = 368
    First_Ams_Keyword = 369
    Across = 369
    Break = 370
    Limit = 371
    Nature = 372
    Noise = 373
    Procedural = 374
    Quantity = 375
    Reference = 376
    Spectrum = 377
    Subnature = 378
    Terminal = 379
    Through = 380
    Tolerance = 381
    Last_AMS_Vhdl = 381
    Last_Keyword = 381
    First_Verilog = 382
    Always = 382
    Assign = 383
    Buf = 384
    Bufif0 = 385
    Bufif1 = 386
    Casex = 387
    Casez = 388
    Cmos = 389
    Deassign = 390
    Defparam = 391
    Disable = 392
    Edge = 393
    Endcase = 394
    Endfunction = 395
    Endmodule = 396
    Endprimitive = 397
    Endspecify = 398
    Endtable = 399
    Endtask = 400
    Forever = 401
    Fork = 402
    Highz0 = 403
    Highz1 = 404
    Ifnone = 405
    Initial = 406
    Input = 407
    Join = 408
    Large = 409
    Macromodule = 410
    Medium = 411
    Module = 412
    Negedge = 413
    Nmos = 414
    Notif0 = 415
    Notif1 = 416
    Output = 417
    Pmos = 418
    Posedge = 419
    Primitive = 420
    Pull0 = 421
    Pull1 = 422
    Pulldown = 423
    Pullup = 424
    Realtime = 425
    Reg = 426
    Repeat = 427
    Rcmos = 428
    Rnmos = 429
    Rpmos = 430
    Rtran = 431
    Rtranif0 = 432
    Rtranif1 = 433
    Scalared = 434
    Small = 435
    Specify = 436
    Specparam = 437
    Strong0 = 438
    Strong1 = 439
    Supply0 = 440
    Supply1 = 441
    Tablex = 442
    Task = 443
    Tran = 444
    Tranif0 = 445
    Tranif1 = 446
    Tri = 447
    Tri0 = 448
    Tri1 = 449
    Triand = 450
    Trior = 451
    Trireg = 452
    Vectored = 453
    Wand = 454
    Weak0 = 455
    Weak1 = 456
    Wire = 457
    Wor = 458
    Last_Verilog = 458
    First_V2001 = 459
    Automatic = 459
    Endgenerate = 460
    Genvar = 461
    Localparam = 462
    Unsigned = 463
    Signed = 464
    Last_V2001 = 464
    Uwire = 465
    First_SV3_0 = 466
    Always_Comb = 466
    Always_Ff = 467
    Always_Latch = 468
    Bit = 469
    Byte = 470
    Changed = 471
    Char = 472
    Const = 473
    Continue = 474
    Do = 475
    Endinterface = 476
    Endtransition = 477
    Enum = 478
    Export = 479
    Extern = 480
    Forkjoin = 481
    Iff = 482
    Import = 483
    Int = 484
    Interface = 485
    Logic = 486
    Longint = 487
    Longreal = 488
    Modport = 489
    Packed = 490
    Priority = 491
    Shortint = 492
    Shortreal = 493
    Static = 494
    Struct = 495
    Timeprecision = 496
    Timeunit = 497
    Transition = 498
    Typedef = 499
    Union = 500
    Unique = 501
    Unique0 = 502
    Void = 503
    Last_SV3_0 = 503
    First_SV3_1 = 504
    Chandle = 504
    Class = 505
    Clocking = 506
    Constraint = 507
    Dist = 508
    Endclass = 509
    Endclocking = 510
    Endprogram = 511
    Endproperty = 512
    Endsequence = 513
    Extends = 514
    Final = 515
    First_Match = 516
    Inside = 517
    Intersect = 518
    Join_Any = 519
    Join_None = 520
    Local = 521
    Program = 522
    Rand = 523
    Randc = 524
    Ref = 525
    Solve = 526
    String = 527
    Super = 528
    This = 529
    Throughout = 530
    Var = 531
    Virtual = 532
    Wait_Order = 533
    Last_SV3_1 = 533
    First_SV3_1a = 534
    Covergroup = 534
    Coverpoint = 535
    Endgroup = 536
    Endpackage = 537
    Expect = 538
    Foreach = 539
    Ignore_Bins = 540
    Illegal_Bins = 541
    Matches = 542
    Randcase = 543
    Randsequence = 544
    Tagged = 545
    Wildcard = 546
    Last_SV3_1a = 546
    First_SV2009 = 547
    Implies = 547
    S_Until = 548
    S_Until_With = 549
    Until_With = 550
    Last_SV2009 = 550
    First_Operator = 551
    Op_Equality = 551
    Op_Inequality = 552
    Op_Less = 553
    Op_Less_Equal = 554
    Op_Greater = 555
    Op_Greater_Equal = 556
    Op_Plus = 557
    Op_Minus = 558
    Op_Mul = 559
    Op_Div = 560
    Op_Exp = 561
    Op_Concatenation = 562
    Op_Condition = 563
    Op_Match_Equality = 564
    Op_Match_Inequality = 565
    Op_Match_Less = 566
    Op_Match_Less_Equal = 567
    Op_Match_Greater = 568
    Op_Match_Greater_Equal = 569
    Last_Operator = 569
    First_Attribute = 570
    Base = 570
    Left = 571
    Right = 572
    High = 573
    Low = 574
    Pos = 575
    Val = 576
    Succ = 577
    Pred = 578
    Leftof = 579
    Rightof = 580
    Reverse_Range = 581
    Length = 582
    Delayed = 583
    Stable = 584
    Quiet = 585
    Transaction = 586
    Event = 587
    Active = 588
    Last_Event = 589
    Last_Active = 590
    Last_Value = 591
    Last_Attribute = 591
    First_Vhdl87_Attribute = 592
    Behavior = 592
    Structure = 593
    Last_Vhdl87_Attribute = 593
    First_Vhdl93_Attribute = 594
    Ascending = 594
    Image = 595
    Value = 596
    Driving = 597
    Driving_Value = 598
    Simple_Name = 599
    Instance_Name = 600
    Path_Name = 601
    Last_Vhdl93_Attribute = 601
    First_Vhdl08_Attribute = 602
    Element = 602
    Last_Vhdl08_Attribute = 602
    First_AMS_Attribute = 603
    Contribution = 603
    Dot = 604
    Integ = 605
    Above = 606
    Zoh = 607
    Ltf = 608
    Ztf = 609
    Ramp = 610
    Slew = 611
    Last_AMS_Attribute = 611
    First_Standard = 612
    Std = 612
    Standard = 613
    Boolean = 614
    NFalse = 615
    NTrue = 616
    Character = 617
    Severity_Level = 618
    Note = 619
    Warning = 620
    Error = 621
    Failure = 622
    Universal_Integer = 623
    Universal_Real = 624
    Convertible_Integer = 625
    Convertible_Real = 626
    Integer = 627
    Real = 628
    Time = 629
    Fs = 630
    Ps = 631
    Ns = 632
    Us = 633
    Ms = 634
    Sec = 635
    Min = 636
    Hr = 637
    Max = 638
    Delay_Length = 639
    Now = 640
    Natural = 641
    Positive = 642
    Bit_Vector = 643
    File_Open_Kind = 644
    Read_Mode = 645
    Write_Mode = 646
    Append_Mode = 647
    File_Open_Status = 648
    Open_Ok = 649
    Status_Error = 650
    Name_Error = 651
    Mode_Error = 652
    Foreign = 653
    Boolean_Vector = 654
    To_Bstring = 655
    To_Binary_String = 656
    To_Ostring = 657
    To_Octal_String = 658
    To_Hstring = 659
    To_Hex_String = 660
    Integer_Vector = 661
    Real_Vector = 662
    Time_Vector = 663
    Digits = 664
    Format = 665
    Unit = 666
    Domain_Type = 667
    Quiescent_Domain = 668
    Time_Domain = 669
    Frequency_Domain = 670
    Domain = 671
    Frequency = 672
    Last_Standard = 672
    First_Charname = 673
    Nul = 673
    Soh = 674
    Stx = 675
    Etx = 676
    Eot = 677
    Enq = 678
    Ack = 679
    Bel = 680
    Bs = 681
    Ht = 682
    Lf = 683
    Vt = 684
    Ff = 685
    Cr = 686
    So = 687
    Si = 688
    Dle = 689
    Dc1 = 690
    Dc2 = 691
    Dc3 = 692
    Dc4 = 693
    Nak = 694
    Syn = 695
    Etb = 696
    Can = 697
    Em = 698
    Sub = 699
    Esc = 700
    Fsp = 701
    Gsp = 702
    Rsp = 703
    Usp = 704
    Del = 705
    C128 = 706
    C129 = 707
    C130 = 708
    C131 = 709
    C132 = 710
    C133 = 711
    C134 = 712
    C135 = 713
    C136 = 714
    C137 = 715
    C138 = 716
    C139 = 717
    C140 = 718
    C141 = 719
    C142 = 720
    C143 = 721
    C144 = 722
    C145 = 723
    C146 = 724
    C147 = 725
    C148 = 726
    C149 = 727
    C150 = 728
    C151 = 729
    C152 = 730
    C153 = 731
    C154 = 732
    C155 = 733
    C156 = 734
    C157 = 735
    C158 = 736
    C159 = 737
    Last_Charname = 737
    First_Misc = 738
    Guard = 738
    Deallocate = 739
    File_Open = 740
    File_Close = 741
    Read = 742
    Write = 743
    Flush = 744
    Endfile = 745
    I = 746
    J = 747
    F = 748
    L = 749
    P = 750
    R = 751
    S = 752
    V = 753
    External_Name = 754
    Open_Kind = 755
    First = 756
    Last = 757
    Textio = 758
    Work = 759
    Text = 760
    To_String = 761
    Minimum = 762
    Maximum = 763
    Untruncated_Text_Read = 764
    Textio_Read_Real = 765
    Textio_Write_Real = 766
    Get_Resolution_Limit = 767
    Control_Simulation = 768
    Step = 769
    Index = 770
    Item = 771
    Uu_File_Uu = 772
    Uu_Line_Uu = 773
    Label_Applies_To = 774
    Return_Port_Name = 775
    Map_To_Operator = 776
    Type_Function = 777
    Built_In = 778
    NNone = 779
    Last_Misc = 779
    First_Ieee_Pkg = 780
    Ieee = 780
    Std_Logic_1164 = 781
    VITAL_Timing = 782
    Numeric_Std = 783
    Numeric_Bit = 784
    Numeric_Std_Unsigned = 785
    Std_Logic_Arith = 786
    Std_Logic_Signed = 787
    Std_Logic_Unsigned = 788
    Std_Logic_Textio = 789
    Std_Logic_Misc = 790
    Math_Real = 791
    Last_Ieee_Pkg = 791
    First_Ieee_Name = 792
    Std_Ulogic = 792
    Std_Ulogic_Vector = 793
    Std_Logic = 794
    Std_Logic_Vector = 795
    Rising_Edge = 796
    Falling_Edge = 797
    VITAL_Level0 = 798
    VITAL_Level1 = 799
    Unresolved_Unsigned = 800
    Unresolved_Signed = 801
    To_Integer = 802
    To_Unsigned = 803
    To_Signed = 804
    Resize = 805
    Std_Match = 806
    Shift_Left = 807
    Shift_Right = 808
    Rotate_Left = 809
    Rotate_Right = 810
    To_Bit = 811
    To_Bitvector = 812
    To_Stdulogic = 813
    To_Stdlogicvector = 814
    To_Stdulogicvector = 815
    Is_X = 816
    To_01 = 817
    To_X01 = 818
    To_X01Z = 819
    To_UX01 = 820
    Conv_Signed = 821
    Conv_Unsigned = 822
    Conv_Integer = 823
    Conv_Std_Logic_Vector = 824
    And_Reduce = 825
    Nand_Reduce = 826
    Or_Reduce = 827
    Nor_Reduce = 828
    Xor_Reduce = 829
    Xnor_Reduce = 830
    Ceil = 831
    Floor = 832
    Round = 833
    Log2 = 834
    Sin = 835
    Cos = 836
    Shl = 837
    Shr = 838
    Ext = 839
    Sxt = 840
    Find_Leftmost = 841
    Find_Rightmost = 842
    Last_Ieee_Name = 842
    First_Synthesis = 843
    Allconst = 843
    Allseq = 844
    Anyconst = 845
    Anyseq = 846
    Gclk = 847
    Last_Synthesis = 847
    First_Directive = 848
    Define = 848
    Endif = 849
    Ifdef = 850
    Ifndef = 851
    Include = 852
    Timescale = 853
    Undef = 854
    Protect = 855
    Begin_Protected = 856
    End_Protected = 857
    Key_Block = 858
    Data_Block = 859
    Line = 860
    Celldefine = 861
    Endcelldefine = 862
    Default_Nettype = 863
    Resetall = 864
    Last_Directive = 864
    First_Systask = 865
    Bits = 865
    D_Root = 866
    D_Unit = 867
    Last_Systask = 867
    First_SV_Method = 868
    Size = 868
    Insert = 869
    Delete = 870
    Pop_Front = 871
    Pop_Back = 872
    Push_Front = 873
    Push_Back = 874
    Name = 875
    Len = 876
    Substr = 877
    Exists = 878
    Atoi = 879
    Itoa = 880
    Find = 881
    Find_Index = 882
    Find_First = 883
    Find_First_Index = 884
    Find_Last = 885
    Find_Last_Index = 886
    Num = 887
    Randomize = 888
    Pre_Randomize = 889
    Post_Randomize = 890
    Srandom = 891
    Get_Randstate = 892
    Set_Randstate = 893
    Seed = 894
    State = 895
    Last_SV_Method = 895
    First_BSV = 896
    uAction = 896
    uActionValue = 897
    BVI = 898
    uC = 899
    uCF = 900
    uE = 901
    uSB = 902
    uSBR = 903
    Action = 904
    Endaction = 905
    Actionvalue = 906
    Endactionvalue = 907
    Ancestor = 908
    Clocked_By = 909
    Default_Clock = 910
    Default_Reset = 911
    Dependencies = 912
    Deriving = 913
    Determines = 914
    Enable = 915
    Ifc_Inout = 916
    Input_Clock = 917
    Input_Reset = 918
    Instance = 919
    Endinstance = 920
    Let = 921
    Match = 922
    Method = 923
    Endmethod = 924
    Numeric = 925
    Output_Clock = 926
    Output_Reset = 927
    Par = 928
    Endpar = 929
    Path = 930
    Provisos = 931
    Ready = 932
    Reset_By = 933
    Rule = 934
    Endrule = 935
    Rules = 936
    Endrules = 937
    Same_Family = 938
    Schedule = 939
    Seq = 940
    Endseq = 941
    Typeclass = 942
    Endtypeclass = 943
    Valueof = 944
    uValueof = 945
    Last_BSV = 945
    First_Comment = 946
    Psl = 946
    Pragma = 947
    Synthesis = 948
    Synopsys = 949
    Translate_Off = 950
    Translate_On = 951
    Translate = 952
    Synthesis_Off = 953
    Synthesis_On = 954
    Off = 955
    Last_Comment = 955
    First_PSL = 956
    A = 956
    Af = 957
    Ag = 958
    Ax = 959
    Abort = 960
    Assume_Guarantee = 961
    Before = 962
    Clock = 963
    E = 964
    Ef = 965
    Eg = 966
    Ex = 967
    Endpoint = 968
    Eventually = 969
    Fairness = 970
    Fell = 971
    Forall = 972
    G = 973
    Inf = 974
    Inherit = 975
    Never = 976
    Next_A = 977
    Next_E = 978
    Next_Event = 979
    Next_Event_A = 980
    Next_Event_E = 981
    Prev = 982
    Rose = 983
    Strong = 984
    W = 985
    Whilenot = 986
    Within = 987
    X = 988
    Last_PSL = 988
    First_Edif = 989
    Celltype = 999
    View = 1000
    Viewtype = 1001
    Direction = 1002
    Contents = 1003
    Net = 1004
    Viewref = 1005
    Cellref = 1006
    Libraryref = 1007
    Portinstance = 1008
    Joined = 1009
    Portref = 1010
    Instanceref = 1011
    Design = 1012
    Designator = 1013
    Owner = 1014
    Member = 1015
    Number = 1016
    Rename = 1017
    Userdata = 1018
    Last_Edif = 1018



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
#  Authors:
#    Tristan Gingold
#    Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from ctypes import c_int32

from pydecor import export

from pyGHDL.libghdl import libghdl

__all__ = [
	'Library_Location',
	'Work_Library'
]

from pyGHDL.libghdl._types import NameId


Library_Location = c_int32.in_dll(libghdl, "libraries__library_location")   #: A location for library declarations (such as library WORK). Type ``Location_Type``. Use ``.value`` to access this variable inside libghdl
Work_Library = c_int32.in_dll(libghdl, "libraries__work_library")           #: Library declaration for the work library. Note: the identifier of the work_library is ``work_library_name``, which may be different from 'WORK'. Type: ``Iir_Library_Declaration``. Use ``.value`` to access this variable inside libghdl


@export
def Get_Libraries_Chain():
	"""
	Get the chain of libraries.  Can be used only to read (it mustn't be modified).

	:return: Type ``Iir_Library_Declaration``
	"""
	return libghdl.libraries__get_libraries_chain()


@export
def Add_Design_Unit_Into_Library(Unit, Keep_Obsolete: bool = False) -> None:
	"""
	Add or replace an design unit in the work library. DECL must not have a chain
	(because it may be modified).

	If the design_file of UNIT is not already in the library, a new one is created.

	Units are always appended to the design_file. Therefore, the order is kept.

	:param Unit:          Type: ``Iir_Design_Unit``
	:param Keep_Obsolete: If :obj:`Keep_Obsolete` is True, obsoleted units are
	                      kept in the library.

	                      This is used when a whole design file has to be added
	                      in the library and then processed (without that feature,
	                      redefined units would disappear).
	"""
	libghdl.libraries__add_design_unit_into_library(Unit, Keep_Obsolete)


@export
def Purge_Design_File(Design_File) -> None:
	"""
	Remove the same file as DESIGN_FILE from work library and all of its units.

	:param Design_File: Type: ``Iir_Design_File``
	"""
	libghdl.libraries__purge_design_file(Design_File)


@export
def Find_Entity_For_Component(Name: NameId):
	"""
	Find an entity whose name is :obj:`Name` in any library. |br|
	If there is no such entity, return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`. |br|
	If there are several entities, return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`;

	:param Name: Entity name to search for.
	:return:     Type: ``Iir_Design_Unit``
	"""
	return libghdl.libraries__find_entity_for_component(Name)


@export
def Get_Library_No_Create(Ident: NameId):
	"""
	Get the library named :obj:`Ident`.

	:param Ident: Libryr to look for.
	:return:      Return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir` if it doesn't exist. Type ``Iir_Library_Declaration``
	"""
	return libghdl.libraries__get_library_no_create(Ident)


@export
def Find_Primary_Unit(Library, Name: NameId):
	"""
	Just return the design_unit for :obj:`Name`, or ``NULL`` if not found.

	:param Library: Library to look in. Type: ``Iir_Library_Declaration``
	:param Name:    Primary unit to search for.
	:return:        Type: ``Iir_Design_Unit``
	"""
	return libghdl.libraries__find_primary_unit(Library, Name)



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
# Authors:
#    Tristan Gingold
#    Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from pydecor import export

from pyGHDL.libghdl import libghdl, Iir


@export
def Strip_Denoting_Name(Name: Iir) -> Iir:
	"""
	If :obj:`Name` is a simple or an expanded name, return the denoted declaration.
	Otherwise, return :obj:`Name`.

	:param Name: Simple or an expanded name.
	:return:     Denoted declaration.
	"""
	return libghdl.vhdl__utils__strip_denoting_name(Name)


@export
def Get_Entity(Decl: Iir) -> Iir:
	"""
	This is a wrapper around ``Get_Entity_Name`` to return the entity declaration
	of the entity name of :obj:`Decl`, or ``Null_Iir`` in case of error.

	:param Decl: Declaration
	:return:     Entity
	"""
	return libghdl.vhdl__utils__get_entity(Decl)


@export
def Is_Second_Subprogram_Specification(Spec: Iir) -> bool:
	"""
	Check if :obj:`Spec` is the subprogram specification of a subprogram body
	which was previously declared. In that case, the only use of :obj:`Spec`
	is to match the body with its declaration.

	:param Spec: Specification
	:return:     ``True`` if subprogram specification and previously declared subprogram body match
	"""
	return libghdl.vhdl__utils__is_second_subprogram_specification(Spec)


@export
def Get_Entity_From_Entity_Aspect(Aspect: Iir) -> Iir:
	"""
	Extract the entity from :obj:`Aspect`.

	If :obj:`Aspect` is a component declaration, return :obj:`Aspect`. If it's
	open, return ``Null_Iir``

	:param Aspect: Aspect
	:return:       Entity
	"""
	return libghdl.vhdl__utils__get_entity_from_entity_aspect(Aspect)


@export
def Get_Interface_Of_Formal(Formal: Iir) -> Iir:
	"""
	Get the interface corresponding to the formal name :obj:`Formal`. This is
	always an interface, even if the formal is a name.

	:param Formal: The formal.
	:return:       The corresponding interface.
	"""
	return libghdl.vhdl__utils__get_interface_of_formal(Formal)



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
#  Authors:
#    Tristan Gingold
#
# Package package:  Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2020 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from ctypes import c_int

from pyGHDL.libghdl import libghdl


__all__ = [
	'Std_Logic_1164_Pkg',
	'Std_Logic_Type',
	'Std_Logic_Vector_Type'
]


Std_Logic_1164_Pkg = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__std_logic_1164_pkg"
)

# Get value
Std_Logic_Type = c_int.in_dll(libghdl, "vhdl__ieee__std_logic_1164__std_logic_type")

# Get value
Std_Logic_Vector_Type = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__std_logic_vector_type"
)

# Get value
# Rising_Edge = c_int.in_dll(libghdl, "vhdl__ieee__std_logic_1164__rising_edge")

# Get value
# Falling_Edge = c_int.in_dll(libghdl, "vhdl__ieee__std_logic_1164__falling_edge")



import logging
import pyGHDL.libghdl.vhdl.nodes as nodes
import pyGHDL.libghdl.vhdl.nodes_meta as nodes_meta
import pyGHDL.libghdl.name_table as name_table
import pyGHDL.libghdl.utils as pyutils

log = logging.getLogger(__name__)


def find_def_chain(first, loc):
    n1 = first
    while n1 != nodes.Null_Iir:
        res = find_def(n1, loc)
        if res is not None:
            return res
        n1 = nodes.Get_Chain(n1)
    return None


def find_def(n, loc):
    "Return the node at location :param loc:, or None if not under :param n:"
    if n == nodes.Null_Iir:
        return None
    k = nodes.Get_Kind(n)
    if k in [
        nodes.Iir_Kind.Simple_Name,
        nodes.Iir_Kind.Character_Literal,
        nodes.Iir_Kind.Operator_Symbol,
        nodes.Iir_Kind.Selected_Name,
        nodes.Iir_Kind.Attribute_Name,
        nodes.Iir_Kind.Selected_Element,
    ]:
        n_loc = nodes.Get_Location(n)
        if loc >= n_loc:
            ident = nodes.Get_Identifier(n)
            id_len = name_table.Get_Name_Length(ident)
            if loc < n_loc + id_len:
                return n
        if k == nodes.Iir_Kind.Simple_Name:
            return None
    elif k == nodes.Iir_Kind.Design_File:
        return find_def_chain(nodes.Get_First_Design_Unit(n), loc)
    elif k == nodes.Iir_Kind.Design_Unit:
        # if loc > elocations.Get_End_Location(unit):
        #    return None
        res = find_def_chain(nodes.Get_Context_Items(n), loc)
        if res is not None:
            return res
        unit = nodes.Get_Library_Unit(n)
        return find_def(unit, loc)

    # This is *much* faster than using node_iter!
    for f in pyutils.fields_iter(n):
        typ = nodes_meta.get_field_type(f)
        if typ == nodes_meta.types.Iir:
            attr = nodes_meta.get_field_attribute(f)
            if attr == nodes_meta.Attr.ANone:
                res = find_def(nodes_meta.Get_Iir(n, f), loc)
                if res is not None:
                    return res
            elif attr == nodes_meta.Attr.Chain:
                res = find_def_chain(nodes_meta.Get_Iir(n, f), loc)
                if res is not None:
                    return res
            elif attr == nodes_meta.Attr.Maybe_Ref:
                if not nodes.Get_Is_Ref(n):
                    res = find_def(nodes_meta.Get_Iir(n, f), loc)
                    if res is not None:
                        return res
        elif typ == nodes_meta.types.Iir_List:
            attr = nodes_meta.get_field_attribute(f)
            if attr == nodes_meta.Attr.ANone:
                for n1 in pyutils.list_iter(nodes_meta.Get_Iir_List(n, f)):
                    res = find_def(n1, loc)
                    if res is not None:
                        return res
        elif typ == nodes_meta.types.Iir_Flist:
            attr = nodes_meta.get_field_attribute(f)
            if attr == nodes_meta.Attr.ANone:
                for n1 in pyutils.flist_iter(nodes_meta.Get_Iir_Flist(n, f)):
                    res = find_def(n1, loc)
                    if res is not None:
                        return res

    return None


def goto_definition(n, loc):
    "Return the declaration (as a node) under :param loc: or None"
    ref = find_def(n, loc)
    log.debug("for loc %u found node %s", loc, ref)
    if ref is None:
        return None
    log.debug(
        "for loc %u id=%s",
        loc,
        name_table.Get_Name_Ptr(nodes.Get_Identifier(ref)),
    )
    ent = nodes.Get_Named_Entity(ref)
    return None if ent == nodes.Null_Iir else ent



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
#  Authors:
#    Tristan Gingold
#    Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from ctypes import c_int32, c_bool, POINTER, Structure

from pydecor import export

from pyGHDL.libghdl import libghdl


@export
class Iterator(Structure):
	_fields_ = [
		("chunk", c_int32),
		("chunk_idx", c_int32),
		("remain", c_int32)
	]


@export
def Iterate(List) -> Iterator:
	"""
	Create an iterator for a given list.

	The idiomatic way to iterate is:

	.. code-block:: Python

	   It = Iterate(List)
	   while Is_Valid(It):
	     El = Get_Element(It)
	       # ...
	     Next(It)

	:param List: List to create an iterator from.
	:return:     Iterator structure.
	"""
	func = libghdl.vhdl__lists__iterate
	func.argstype = [c_int32]
	func.restype = Iterator

	return func(List)


@export
def Is_Valid(it: Iterator) -> bool:
	"""
	Check if iterator reached the end.

	:param Iterator: Iterator to check.
	:return:         False, if iterator has reached the end.
	"""
	func = libghdl.vhdl__lists__is_valid
	func.argstype = [POINTER(Iterator)]
	func.restype = c_bool

	return func(it)


@export
def Next(it: Iterator):
	"""
	Move iterator to the next element.

	:param Iterator: Iterator to increment.
	:return:         False, if iterator has reached the end.
	"""
	func = libghdl.vhdl__lists__next
	func.argstype = [POINTER(Iterator)]
	func.restype = None

	func(it)


@export
def Get_Element(it: Iterator) -> int:
	"""
	Get the current element from iterator.

	:param Iterator: Iterator the get the element from.
	:return:         The current element the iterator points to. Type: ``El_Type``
	"""
	func = libghdl.vhdl__lists__get_element
	func.argstype = [POINTER(Iterator)]
	func.restype = c_int32

	return func(it)


@export
def Get_Nbr_Elements(List) -> int:
	"""
	Return the number of elements in the list.

	.. hint:: This is also 1 + the position of the last element.

	:param List: The list to use.
	:return:     Number of list elements.
	"""
	func = libghdl.vhdl__lists__get_nbr_elements
	func.argtype = [c_int32]
	func.restype = c_int32

	return func(List)


@export
def Create_Iir_List() -> int:
	"""
	Create a list.

	:return: Type: ``List_Type``
	"""
	return libghdl.vhdl__lists__create_list()


@export
def Destroy_Iir_List(List) -> None:
	"""
	Destroy a list.

	:param List: List to destroy.
	"""
	libghdl.vhdl__lists__destroy_list(List)



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
#  Authors:
#    Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from typing import TypeVar

__all__ = [
	'ErrorIndex',
	'MessageIdWarnings',
	'NameId',
	'SourceFileEntry',
	'Iir',
	'IirKind'
]

ErrorIndex =  TypeVar('ErrorIndex', bound=int)
MessageIdWarnings = TypeVar('MessageIdWarnings', bound=int)
NameId = TypeVar('NameId', bound=int)
SourceFileEntry = TypeVar('SourceFileEntry', bound=int)

Iir = TypeVar('Iir', bound=int)
IirKind = TypeVar('IirKind', bound=int)



#!/usr/bin/env python
# =============================================================================
#                ____ _   _ ____  _           _ _
#   _ __  _   _ / ___| | | |  _ \| |      ___| (_)
#  | '_ \| | | | |  _| |_| | | | | |     / __| | |
#  | |_) | |_| | |_| |  _  | |_| | |___ | (__| | |
#  | .__/ \__, |\____|_| |_|____/|_____(_)___|_|_|
#  |_|    |___/
# =============================================================================
#  Authors:
#   Tristan Gingold
#
# Package module:   GHDLs Language Server implementing LSP for VHDL.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2020 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from __future__ import absolute_import

from argparse import ArgumentParser
from logging import getLogger, DEBUG, INFO, ERROR, basicConfig
import sys
import os

from pydecor import export

import pyGHDL.libghdl as libghdl
from pyGHDL.libghdl     import version, errorout_console
from pyGHDL.lsp         import LSPConnTrace
from pyGHDL.lsp.lsp     import LSPConn, LanguageProtocolServer
from pyGHDL.lsp.vhdl_ls import VhdlLanguageServer

__loggerName = "ghdl-ls"


def __rotate_log_files(basename: str, num: int):
    """Rotate existing log files."""
    for i in range(num, 0, -1):
        oldfile = "{}.{}".format(basename, i - 1)
        if os.path.isfile(oldfile):
            os.rename(oldfile, "{}.{}".format(basename, i))
    if os.path.isfile(basename):
        os.rename(basename, "{}.0".format(basename))


def _generateCLIParser() -> ArgumentParser:
    """Creates an CLI argument parser based on ``argparse``."""
    parser = ArgumentParser(description="VHDL Language Protocol Server. Find info about clients in `ghdl/ghdl-language-server <https://github.com/ghdl/ghdl-language-server>`__.")
    parser.add_argument(
        "--version", "-V", action="version", version="%(prog)s " + version.__version__
    )
    parser.add_argument(
        "--verbose", "-v", action="count", default=0, help="Show debug output"
    )
    parser.add_argument(
        "--log-file", help="Redirect logs to the given file instead of stderr"
    )
    parser.add_argument("--trace-file", help="Save RPC data to FILE.in and FILE.out (overrides :envvar:`GHDL_LS_TRACE`)")
    parser.add_argument("--input", "-i", help="Read request from file")
    parser.add_argument(
        "--disp-config",
        action="store_true",
        help="Display installation configuration and exit",
    )

    return parser


@export
def main():
    """Entrypoint of GHDL's Language Protocol Server."""
    logger = getLogger(__loggerName)

    parser = _generateCLIParser()
    args = parser.parse_args()

    if args.disp_config:
        errorout_console.Install_Handler()
        libghdl.disp_config()
        return

    # Setup logging
    if args.verbose >= 2:
        loglevel = DEBUG
    elif args.verbose >= 1:
        loglevel = INFO
    else:
        loglevel = ERROR

    if args.log_file:
        __rotate_log_files(args.log_file, 5)
        logstream = open(args.log_file, "w")
    else:
        logstream = sys.stderr

    basicConfig(
        format="%(asctime)-15s [%(levelname)s] %(message)s",
        stream=logstream,
        level=loglevel,
    )

    if args.verbose != 0:
        sys.stderr.write("Args: {}\n".format(sys.argv))
        sys.stderr.write("Current directory: {}\n".format(os.getcwd()))

    logger.info("Args: %s", sys.argv)
    logger.info("Current directory is %s", os.getcwd())

    # Connection
    instream = sys.stdin.buffer
    if args.input is not None:
        instream = open(args.input, "rb")

    conn = LSPConn(instream, sys.stdout.buffer)

    trace_file = args.trace_file
    if trace_file is None:
        trace_file = os.environ.get("GHDL_LS_TRACE")
    if trace_file is not None:
        if args.input is None:
            __rotate_log_files(trace_file + ".in", 5)
            __rotate_log_files(trace_file + ".out", 5)
            conn = LSPConnTrace(trace_file, conn)
        else:
            logger.info("Traces disabled when -i/--input")

    handler = VhdlLanguageServer()

    try:
        server = LanguageProtocolServer(handler, conn)
        server.run()
    except Exception:
        logger.exception("Uncaught error")
        sys.exit(1)


if __name__ == '__main__':
    main()



# Auto generated Python source file from Ada sources
# Call 'make' in 'src/vhdl' to regenerate:
#
from pydecor import export
from pyGHDL.libghdl import libghdl

Get_Start_Location = libghdl.vhdl__elocations__get_start_location
Set_Start_Location = libghdl.vhdl__elocations__set_start_location

Get_Right_Paren_Location = libghdl.vhdl__elocations__get_right_paren_location
Set_Right_Paren_Location = libghdl.vhdl__elocations__set_right_paren_location

Get_End_Location = libghdl.vhdl__elocations__get_end_location
Set_End_Location = libghdl.vhdl__elocations__set_end_location

Get_Is_Location = libghdl.vhdl__elocations__get_is_location
Set_Is_Location = libghdl.vhdl__elocations__set_is_location

Get_Begin_Location = libghdl.vhdl__elocations__get_begin_location
Set_Begin_Location = libghdl.vhdl__elocations__set_begin_location

Get_Then_Location = libghdl.vhdl__elocations__get_then_location
Set_Then_Location = libghdl.vhdl__elocations__set_then_location

Get_Use_Location = libghdl.vhdl__elocations__get_use_location
Set_Use_Location = libghdl.vhdl__elocations__set_use_location

Get_Loop_Location = libghdl.vhdl__elocations__get_loop_location
Set_Loop_Location = libghdl.vhdl__elocations__set_loop_location

Get_Generate_Location = libghdl.vhdl__elocations__get_generate_location
Set_Generate_Location = libghdl.vhdl__elocations__set_generate_location

Get_Generic_Location = libghdl.vhdl__elocations__get_generic_location
Set_Generic_Location = libghdl.vhdl__elocations__set_generic_location

Get_Port_Location = libghdl.vhdl__elocations__get_port_location
Set_Port_Location = libghdl.vhdl__elocations__set_port_location

Get_Generic_Map_Location = libghdl.vhdl__elocations__get_generic_map_location
Set_Generic_Map_Location = libghdl.vhdl__elocations__set_generic_map_location

Get_Port_Map_Location = libghdl.vhdl__elocations__get_port_map_location
Set_Port_Map_Location = libghdl.vhdl__elocations__set_port_map_location

Get_Arrow_Location = libghdl.vhdl__elocations__get_arrow_location
Set_Arrow_Location = libghdl.vhdl__elocations__set_arrow_location

Get_Colon_Location = libghdl.vhdl__elocations__get_colon_location
Set_Colon_Location = libghdl.vhdl__elocations__set_colon_location

Get_Assign_Location = libghdl.vhdl__elocations__get_assign_location
Set_Assign_Location = libghdl.vhdl__elocations__set_assign_location




import os
import logging
import json
from urllib.parse import unquote, quote

log = logging.getLogger("ghdl-ls")

is_windows = os.name == "nt"

class ProtocolError(Exception):
    pass


class LSPConn:
    def __init__(self, reader, writer):
        self.reader = reader
        self.writer = writer

    def readline(self):
        data = self.reader.readline()
        return data.decode("utf-8")

    def read(self, size):
        data = self.reader.read(size)
        return data.decode("utf-8")

    def write(self, out):
        self.writer.write(out.encode())
        self.writer.flush()


def path_from_uri(uri):
    # Convert file uri to path (strip html like head part)
    # This is needed to get the root path and to load a document when the
    # textual source is not present.
    if not uri.startswith("file://"):
        # No scheme
        return uri
    _, path = uri.split("file://", 1)
    if is_windows and path.startswith("/"):
        # On windows, absolute files start like "/C:/aa/bbb".
        # Remove the first "/".
        path = path[1:]
    return os.path.normpath(unquote(path))


def path_to_uri(path):
    # Convert path to file uri (add html like head part)
    if is_windows:
        return "file:///" + quote(path.replace("\\", "/"))
    else:
        return "file://" + quote(path)


class LanguageProtocolServer(object):
    def __init__(self, handler, conn):
        self.conn = conn
        self.handler = handler
        if handler is not None:
            handler.set_lsp(self)
        self.running = True
        self._next_id = 0

    def read_request(self):
        headers = {}
        while True:
            # Read a line
            line = self.conn.readline()
            # Return on EOF.
            if not line:
                return None
            if line[-2:] != "\r\n":
                raise ProtocolError("invalid end of line in header")
            line = line[:-2]
            if not line:
                # End of headers.
                log.debug("Headers: %r", headers)
                length = headers.get("Content-Length", None)
                if length is not None:
                    body = self.conn.read(int(length))
                    return body
                else:
                    raise ProtocolError("missing Content-Length in header")
            else:
                key, value = line.split(": ", 1)
                headers[key] = value

    def run(self):
        while self.running:
            body = self.read_request()
            if body is None:
                # EOF
                break

            # Text to JSON
            msg = json.loads(body)
            log.debug("Read msg: %s", msg)

            reply = self.handle(msg)
            if reply is not None:
                self.write_output(reply)

    def handle(self, msg):
        if msg.get("jsonrpc", None) != "2.0":
            raise ProtocolError("invalid jsonrpc version")
        tid = msg.get("id", None)
        method = msg.get("method", None)
        if method is None:
            # This is a reply.
            log.error("Unexpected reply for %s", tid)
            return
        params = msg.get("params", None)
        fmethod = self.handler.dispatcher.get(method, None)
        if fmethod:
            if params is None:
                params = {}
            try:
                response = fmethod(**params)
            except Exception as e:
                log.exception(
                    "Caught exception while handling %s with params %s:", method, params
                )
                self.show_message(
                    MessageType.Error,
                    (
                        "Caught exception while handling {}, "
                        + "see VHDL language server output for details."
                    ).format(method),
                )
                response = None
            if tid is None:
                # If this was just a notification, discard it
                return None
            log.debug("Response: %s", response)
            rbody = {
                "jsonrpc": "2.0",
                "id": tid,
                "result": response,
            }
        else:
            # Unknown method.
            log.error("Unknown method %s", method)
            # If this was just a notification, discard it
            if tid is None:
                return None
            # Otherwise create an error.
            rbody = {
                "jsonrpc": "2.0",
                "id": tid,
                "error": {
                    "code": JSONErrorCodes.MethodNotFound,
                    "message": "unknown method {}".format(method),
                },
            }
        return rbody

    def write_output(self, body):
        output = json.dumps(body, separators=(",", ":"))
        self.conn.write("Content-Length: {}\r\n".format(len(output)))
        self.conn.write("\r\n")
        self.conn.write(output)

    def notify(self, method, params):
        """Send a notification"""
        body = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
        }
        self.write_output(body)

    def send_request(self, method, params):
        """Send a request"""
        self._next_id += 1
        body = {
            "jsonrpc": "2.0",
            "id": self._next_id,
            "method": method,
            "params": params,
        }
        self.write_output(body)

    def shutdown(self):
        """Prepare to shutdown the server"""
        self.running = False

    def show_message(self, typ, message):
        self.notify("window/showMessage", {"type": typ, "message": message})

    def configuration(self, items):
        return self.send_request("workspace/configuration", {"items": items})


# ----------------------------------------------------------------------
#  Standard defines and object types
#


class JSONErrorCodes(object):
    # Defined by JSON RPC
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    serverErrorStart = -32099
    serverErrorEnd = -32000
    ServerNotInitialized = -32002
    UnknownErrorCode = -32001

    # Defined by the protocol.
    RequestCancelled = -32800
    ContentModified = -32801


class CompletionKind(object):
    Text = 1
    Method = 2
    Function = 3
    Constructor = 4
    Field = 5
    Variable = 6
    Class = 7
    Interface = 8
    Module = 9
    Property = 10
    Unit = 11
    Value = 12
    Enum = 13
    Keyword = 14
    Snippet = 15
    Color = 16
    File = 17
    Reference = 18


class DiagnosticSeverity(object):
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


class TextDocumentSyncKind(object):
    NONE = (0,)
    FULL = 1
    INCREMENTAL = 2


class MessageType(object):
    Error = 1
    Warning = 2
    Info = 3
    Log = 4


class SymbolKind(object):
    File = 1
    Module = 2
    Namespace = 3
    Package = 4
    Class = 5
    Method = 6
    Property = 7
    Field = 8
    Constructor = 9
    Enum = 10
    Interface = 11
    Function = 12
    Variable = 13
    Constant = 14
    String = 15
    Number = 16
    Boolean = 17
    Array = 18



# =============================================================================
#                ____ _   _ ____  _          _
#   _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
#  | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
#  | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
#  | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
#  |_|    |___/
# =============================================================================
#  Authors:
#    Patrick Lehmann
#
# Package module:   DOM: Elements not covered by the VHDL standard.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2020 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

"""
.. todo::
   Add a module documentation.
"""
from pathlib import Path
from typing  import Any

from pydecor import export

from pyVHDLModel.VHDLModel  import Design        as VHDLModel_Design
from pyVHDLModel.VHDLModel  import Library       as VHDLModel_Library
from pyVHDLModel.VHDLModel  import Document      as VHDLModel_Document

import pyGHDL.libghdl       as libghdl
from pyGHDL.libghdl         import name_table, files_map, errorout_memory
from pyGHDL.libghdl.vhdl    import nodes, sem_lib

from pyGHDL.dom.Common      import LibGHDLException, GHDLException
from pyGHDL.dom.DesignUnit  import Entity, Architecture, Package, PackageBody, Context, Configuration

__all__ = []


@export
class Design(VHDLModel_Design):
	def __init__(self):
		super().__init__()

		self.__ghdl_init()

	def __ghdl_init(self):
		"""Initialization: set options and then load libraries"""
		# Initialize libghdl
		libghdl.finalize()
		libghdl.initialize()

		# Collect error messages in memory
		errorout_memory.Install_Handler()

		libghdl.set_option("--std=08")

		# Finish initialization. This will load the standard package.
		if libghdl.analyze_init_status() != 0:
			raise LibGHDLException("Error initializing 'libghdl'.")


@export
class Library(VHDLModel_Library):
	pass


@export
class Document(VHDLModel_Document):
	__ghdlFileID:           Any
	__ghdlSourceFileEntry:  Any
	__ghdlFile:             Any

	def __init__(self, path : Path = None, dontParse: bool = False):
		super().__init__(path)

		self.__ghdl_init()
		if (dontParse == False):
			self.parse()

	def __ghdl_init(self):
		# Read input file
		self.__ghdlFileID = name_table.Get_Identifier(str(self.Path))
		self.__ghdlSourceFileEntry = files_map.Read_Source_File(name_table.Null_Identifier, self.__ghdlFileID)
		if self.__ghdlSourceFileEntry == files_map.No_Source_File_Entry:
			raise LibGHDLException("Cannot load file '{!s}'".format(self.Path))

		# parse
		self.__ghdlFile =  sem_lib.Load_File(self.__ghdlSourceFileEntry)

	def parse(self):
		unit = nodes.Get_First_Design_Unit(self.__ghdlFile)
		while unit != nodes.Null_Iir:
			libraryUnit = nodes.Get_Library_Unit(unit)
			nodeKind = nodes.Get_Kind(libraryUnit)

			if (nodeKind == nodes.Iir_Kind.Entity_Declaration):
				entity = Entity.parse(libraryUnit)
				self.Entities.append(entity)

			elif (nodeKind == nodes.Iir_Kind.Architecture_Body):
				architecture = Architecture.parse(libraryUnit)
				self.Architectures.append(architecture)

			elif (nodeKind == nodes.Iir_Kind.Package_Declaration):
				package = Package.parse(libraryUnit)
				self.Packages.append(package)

			elif (nodeKind == nodes.Iir_Kind.Package_Body):
				packageBody = PackageBody.parse(libraryUnit)
				self.PackageBodies.append(packageBody)

			elif (nodeKind == nodes.Iir_Kind.Context_Declaration):
				context = Context.parse(libraryUnit)
				self.Contexts.append(context)

			elif (nodeKind == nodes.Iir_Kind.Configuration_Declaration):
				configuration = Configuration.parse(libraryUnit)
				self.Configurations.append(configuration)

			else:
				raise GHDLException("Unknown design unit kind.")

			unit = nodes.Get_Chain(unit)
